package cmd

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"log"
	"os"
	"path"

	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/sergi/go-diff/diffmatchpatch"

	"github.com/spf13/cobra"
	yaml "gopkg.in/yaml.v2"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/xbus/storage"
)

type pipelineData struct {
	api.PipelineInfo `yaml:",inline"`
	storage.Graph    `yaml:",inline"`
	SourceFile       string `yaml:"-"`
	isNewVersion     bool   `yaml:"-"`
	isNewPipeline    bool   `yaml:"-"`
}

type pipelineSet struct {
	data map[string]map[string]pipelineData
}

func newPipelineSet() *pipelineSet {
	return &pipelineSet{
		make(map[string]map[string]pipelineData),
	}
}

func (s *pipelineSet) loadFile(fname string) error {
	b, err := ioutil.ReadFile(fname)
	if err != nil {
		return err
	}

	if bytes.HasPrefix(b, []byte("pipelines:\n")) {
		var data = make(map[string]map[string]map[string]pipelineData)
		if err := yaml.Unmarshal(b, &data); err != nil {
			return err
		}
		if len(data) != 1 {
			return fmt.Errorf("expects a single 'pipelines' entry at the top level")
		}
		for name, versions := range data["pipelines"] {
			pversions, ok := s.data[name]
			if !ok {
				pversions = make(map[string]pipelineData)
				s.data[name] = pversions
			}
			for version, definition := range versions {
				definition.SourceFile = fname
				if definition.Nodes == nil {
					return fmt.Errorf("pipeline %s-%s in %s has no valid 'nodes' field", name, version, fname)
				}
				definition.Name = name
				definition.Version = version
				if d, ok := pversions[version]; ok {
					return fmt.Errorf(
						"file %s redefines the pipeline %s version %s (first defined in %s)",
						fname, name, version, d.SourceFile,
					)
				}
				pversions[version] = definition
				Logger.Debugf("Loaded %s-%s", name, version)
			}
		}
	} else {
		name := fname[:len(fname)-len(path.Ext(fname))]
		name = path.Base(name)

		var data pipelineData
		if err := yaml.Unmarshal(b, &data); err != nil {
			return err
		}
		data.SourceFile = fname
		if data.Nodes == nil {
			return fmt.Errorf("%s file has no valid 'nodes' field", fname)
		}
		data.Name = name
		pversions, ok := s.data[name]
		if !ok {
			pversions = make(map[string]pipelineData)
			s.data[name] = pversions
		}
		if d, ok := pversions[data.Version]; ok {
			return fmt.Errorf(
				"file %s redefines the pipeline %s version %s (first defined in %s)",
				fname, name, data.Version, d.SourceFile,
			)
		}
		pversions[data.Version] = data
		Logger.Debugf("Loaded %s-%s", name, data.Version)
	}

	Logger.Debugf("-> Loaded %s", fname)

	return nil
}

func (s *pipelineSet) loadDir(dirname string) error {
	f, err := os.Open(dirname)
	if err != nil {
		log.Fatal(err)
	}
	files, err := f.Readdir(-1)
	f.Close()
	if err != nil {
		log.Fatal(err)
	}

	for _, file := range files {
		if err := s.loadFile(path.Join(dirname, file.Name())); err != nil {
			return err
		}
	}
	return nil
}

func (s *pipelineSet) Add(info api.PipelineInfo) error {
	pversions, ok := s.data[info.Name]
	if !ok {
		pversions = make(map[string]pipelineData)
		s.data[info.Name] = pversions
	}
	if _, ok := pversions[info.Version]; ok {
		return fmt.Errorf("duplicate entry: %s-%s", info.Name, info.Version)
	}
	pversions[info.Version] = pipelineData{PipelineInfo: info}
	return nil
}

func (s *pipelineSet) SetGraph(info api.PipelineInfo, graph storage.Graph) error {
	pversions, ok := s.data[info.Name]
	if !ok {
		return fmt.Errorf("no such pipeline: %s", info.Name)
	}
	data, ok := pversions[info.Version]
	if !ok {
		return fmt.Errorf("no such pipeline: %s-%s", info.Name, info.Version)
	}
	data.Graph = graph
	pversions[info.Version] = data
	return nil
}

func (s *pipelineSet) GetInfo(name, version string) (api.PipelineInfo, bool) {
	pversions, ok := s.data[name]
	if !ok {
		return api.PipelineInfo{}, false
	}
	data, ok := pversions[version]
	if !ok {
		return api.PipelineInfo{}, false
	}
	return data.PipelineInfo, true
}

func (s *pipelineSet) GetGraph(info api.PipelineInfo) (storage.Graph, error) {
	pversions, ok := s.data[info.Name]
	if !ok {
		return storage.Graph{}, fmt.Errorf("no such pipeline: %s", info.Name)
	}
	data, ok := pversions[info.Version]
	if !ok {
		return storage.Graph{}, fmt.Errorf("no such pipeline: %s-%s", info.Name, info.Version)
	}
	if len(data.Graph.Nodes) == 0 {
		return storage.Graph{}, fmt.Errorf("graph %s-%s is missing", info.Name, info.Version)
	}
	return data.Graph, nil
}

func (s *pipelineSet) HasGraph(info api.PipelineInfo) bool {
	pversions, ok := s.data[info.Name]
	if !ok {
		return false
	}
	data, ok := pversions[info.Version]
	if !ok {
		return false
	}
	if len(data.Graph.Nodes) == 0 {
		return false
	}
	return true
}

func (s *pipelineSet) Range(fn func(pipelineData) error) error {
	for _, versions := range s.data {
		for _, version := range versions {
			if err := fn(version); err != nil {
				return err
			}
		}
	}
	return nil
}

func pipelineSetAll(sources []string, dryRun, yes bool) error {
	pset := newPipelineSet()

	for _, source := range sources {
		info, err := os.Stat(source)
		if err != nil {
			return err
		}
		if info.IsDir() {
			if err := pset.loadDir(source); err != nil {
				return err
			}
		} else {
			if err := pset.loadFile(source); err != nil {
				return err
			}
		}
	}

	activeVersions := map[string]string{}
	if err := pset.Range(func(data pipelineData) error {
		if v, ok := activeVersions[data.Name]; ok {
			return fmt.Errorf(
				"pipeline %s has multiple active versions (which is not possible): %s and %s",
				data.Name, v, data.Version)
		}
		activeVersions[data.Name] = data.Version
		return nil
	}); err != nil {
		return err
	}

	// Now compare our set with the server pipeline set
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	var (
		serverDataSet        = newPipelineSet()
		needSave             []pipelineData
		needSaveDiffs        [][]diffmatchpatch.Diff
		needStatusChangeFrom []api.PipelineInfo
		needStatusChangeTo   []api.PipelineInfo
	)
	infos, err := c.PipelineQuery("")
	if err != nil {
		return err
	}
	for _, info := range infos {
		if err := serverDataSet.Add(info); err != nil {
			return err
		}
	}
	if err := pset.Range(func(data pipelineData) error {
		info, exists := serverDataSet.GetInfo(data.Name, data.Version)
		data.isNewVersion = !exists
		if data.isNewVersion {
			info.Name = data.Name
			info.Version = data.Version
			info.Status = api.PipelineInfo_DRAFT
		}
		if data.Status == api.PipelineInfo_DRAFT && info.Status != api.PipelineInfo_DRAFT {
			return fmt.Errorf(
				"Cannot change %s-%s status from %s to DRAFT",
				data.Name, data.Version, info.Status,
			)
		}
		var graph storage.Graph
		b := bytes.NewBuffer(nil)
		// get the graph from the server
		// TODO check that the server is not a newer version so we are sure
		// to be able to parse the graph
		if !data.isNewVersion {
			if err := pipelineGet(c, data.Name, data.Version, b); err != nil {
				return err
			}
		} else {
			if err := pipelineGet(c, data.Name, "", b); err != nil {
				if ne, ok := err.(*nrpc.Error); ok && ne.Message == "No such graph" {
					data.isNewPipeline = true
				} else {
					return err
				}
			}
		}
		if err := yaml.Unmarshal(b.Bytes(), &graph); err != nil {
			return err
		}

		// Diff the two graphs and fail if different and non-draft
		diffs, err := diffGraph(graph, data.Graph)
		if err != nil {
			return err
		}
		if len(diffs) != 1 || diffs[0].Type != diffmatchpatch.DiffEqual {
			if info.Status != api.PipelineInfo_DRAFT {
				fmt.Println(diffs)
				s := fmt.Sprintf(
					"Non-draft pipeline %s-%s differs from server:",
					info.Name, info.Version)
				fmt.Println(s)
				fmt.Println(diffmatchpatch.New().DiffPrettyText(diffs))
				return fmt.Errorf(s)
			}
			needSave = append(needSave, data)
			needSaveDiffs = append(needSaveDiffs, diffs)
		}
		if data.Status != info.Status {
			needStatusChangeFrom = append(needStatusChangeFrom, info)
			needStatusChangeTo = append(needStatusChangeTo, data.PipelineInfo)
		}
		return nil
	}); err != nil {
		return err
	}

	for i, data := range needSave {
		diffs := needSaveDiffs[i]
		if data.isNewPipeline {
			fmt.Printf("Will save the NEW pipeline %s at its initial version, %s\n", data.Name, data.Version)
		} else if data.isNewVersion {
			fmt.Printf("Will save a NEW version of %s: %s\n", data.Name, data.Version)
		} else {
			fmt.Printf("Will save the new revision of %s-%s\n", data.Name, data.Version)
		}
		fmt.Printf(
			"  Changes:\n%s\n",
			IndentString(diffmatchpatch.New().DiffPrettyText(diffs), "    "),
		)
	}

	for i, to := range needStatusChangeTo {
		from := needStatusChangeFrom[i]
		fmt.Printf(
			"Will change %s-%s state from %s to %s\n",
			to.Name, to.Version, from.Status, to.Status)
	}

	if len(needSave) == 0 && len(needStatusChangeTo) == 0 {
		fmt.Println("Nothing to do")
		return nil
	}

	if !dryRun {
		if !yes && !askForConfirmation("Please confirm") {
			return fmt.Errorf("Operation canceled")
		}

		saveErrors := map[string]bool{}

		for _, data := range needSave {
			//diffs := needSaveDiffs[i]
			fmt.Printf("Saving %s-%s (from %s)\n", data.Name, data.Version, data.SourceFile)

			content, err := yaml.Marshal(data.Graph)
			if err != nil {
				return err
			}

			_, _, validationMessages, err := c.PipelineSave(data.PipelineInfo, string(content))

			if err != nil {
				saveErrors[data.Name+"##"+data.Version] = true

				fmt.Printf("  error: %s\n", err)
			}

			if len(validationMessages) != 0 {
				saveErrors[data.Name+"##"+data.Version] = true

				fmt.Printf("  Validation messages:\n")
				for _, msg := range validationMessages {
					fmt.Println("    -", msg)
				}
			}

		}

		for i, to := range needStatusChangeTo {
			if saveErrors[to.Name+"##"+to.Version] {
				fmt.Printf(
					"SKIP %s-%s status change because it had errors during save\n",
					to.Name, to.Version)
				continue
			}
			from := needStatusChangeFrom[i]
			fmt.Printf(
				"Changing %s-%s state from %s to %s\n",
				to.Name, to.Version, from.Status, to.Status)
			if _, err := c.PipelineSetStatus(to); err != nil {
				fmt.Printf("%s-%s failed to change state. error was: %s\n",
					to.Name, to.Version, err)
				return err
			}
		}
	}

	return nil
}

func diffGraph(g1, g2 storage.Graph) ([]diffmatchpatch.Diff, error) {
	var b []byte
	if len(g1.Nodes) != 0 || len(g1.Edges) != 0 {
		var err error
		b, err = yaml.Marshal(g1)
		if err != nil {
			return nil, err
		}
	}
	text1 := string(b)

	b, err := yaml.Marshal(g2)
	if err != nil {
		return nil, err
	}
	text2 := string(b)

	dmp := diffmatchpatch.New()
	diffs := dmp.DiffMain(text1, text2, false)
	return diffs, nil
}

var pipelineSetAllCmd = &cobra.Command{
	Use:   "set-all <name or id> [version]",
	Short: "Set all the pipeline definitions",
	Long:  "Set all the pipeline definitions from one or multiple files",
	Run: func(cmd *cobra.Command, args []string) {
		var sources = args
		if len(sources) == 0 {
			sources = []string{"."}
		}
		var (
			dryRun = cmd.Flag("dry-run").Value.String() == "true"
			yes    = cmd.Flag("yes").Value.String() == "true"
		)
		if err := pipelineSetAll(sources, dryRun, yes); err != nil {
			Logger.Errorf("%s", err)
			os.Exit(-1)
		}
	},
}

func init() {
	pipelineSetAllCmd.Flags().Bool("dry-run", false, "Display what would be done")
	pipelineSetAllCmd.Flags().Bool("yes", false, "Assume 'yes' when asked any question")
	pipelineCmd.AddCommand(pipelineSetAllCmd)
}
