package logsentry

import (
	"bytes"
	"fmt"
	"io"

	"github.com/getsentry/sentry-go"
	jsoniter "github.com/json-iterator/go"
	"github.com/rs/zerolog"
)

// GetSentryOutputMiddleware returns a zerolog.LevelWriter wrapper that sends
// any message above "warning" level to the given sentry hub
func GetSentryOutputMiddleware(hub *sentry.Hub) func(io.Writer) io.Writer {
	return func(next io.Writer) io.Writer {
		return SentryWriter{hub, next}
	}
}

// SentryWriter ...
type SentryWriter struct {
	hub  *sentry.Hub
	next io.Writer
}

var (
	levelMarker    = []byte(`"level":"`)
	levelMarkerLen = len(levelMarker)
)

func getLevel(p []byte) zerolog.Level {
	if i := bytes.Index(p, levelMarker); i != -1 {
		first := p[i+levelMarkerLen]
		switch first {
		case 't':
			return zerolog.TraceLevel
		case 'd':
			return zerolog.DebugLevel
		case 'i':
			return zerolog.InfoLevel
		case 'w':
			return zerolog.WarnLevel
		case 'e':
			return zerolog.ErrorLevel
		case 'f':
			return zerolog.FatalLevel
		case 'p':
			return zerolog.PanicLevel
		default:
			return zerolog.WarnLevel
		}
	}
	return zerolog.WarnLevel
}

func buildEvent(p []byte) *sentry.Event {
	iter := jsoniter.ConfigFastest.BorrowIterator(p)
	defer jsoniter.ConfigFastest.ReturnIterator(iter)

	if iter.WhatIsNext() != jsoniter.ObjectValue {
		return nil
	}
	event := sentry.NewEvent()

	if !iter.ReadObjectCB(func(iter *jsoniter.Iterator, field string) bool {
		switch field {
		case "message":
			event.Message = iter.ReadString()
		case "error":
			errMsg := iter.ReadString()
			if event.Message == "" {
				event.Message = errMsg
			}
			event.Extra["error"] = errMsg
		case "level":
			level := iter.ReadString()
			if level == "warn" {
				event.Level = sentry.LevelWarning
			} else {
				event.Level = sentry.Level(level)
			}
		case "request":
			iter.ReadVal(&event.Request)
		case "user":
			if !iter.ReadObjectCB(func(iter *jsoniter.Iterator, field string) bool {
				switch field {
				case "id":
					event.User.ID = iter.ReadString()
				default:
					event.Extra["user."+field] = iter.Read()
				}
				return true
			}) {
				return false
			}
		case "exception":
			var e sentry.Exception
			iter.ReadVal(&e)
			event.Exception = []sentry.Exception{e}
		default:
			event.Extra[field] = iter.Read()
		}
		return true
	}) {
		fmt.Println("Error building sentry obj", iter.Error)
		return nil
	}
	if op, ok := event.Extra["api-operation-id"]; ok {
		tag, _ := event.Extra["api-tag"].(string)
		event.Message = tag + "/" + op.(string) + ": " + event.Message
	} else if event.Request != nil && event.Request.Method != "" {
		event.Message = event.Request.Method + " " + event.Request.URL + ": " + event.Message
	}
	return event
}

// Write ...
func (l SentryWriter) Write(p []byte) (n int, err error) {
	return l.WriteLevel(getLevel(p), p)
}

// WriteLevel implements zerolog.LevelWriter interface
func (l SentryWriter) WriteLevel(level zerolog.Level, p []byte) (int, error) {
	if level >= zerolog.WarnLevel {
		if e := buildEvent(p); e != nil {
			l.hub.CaptureEvent(e)
		}
	}

	return l.next.Write(p)
}
